@@ -94,15 +94,52 @@ static gchar *ayuv_to_rgba_shader =
     "  color.b = y + 2.015625 * u;"
     "  cogl_color_out = color;}";
 
+static gchar *nv12_to_rgba_shader =
+    "uniform sampler2D ytex;"
+    "uniform sampler2D utex;"
+    "void main () {"
+    "  vec2 coord = vec2(cogl_tex_coord_in[0]);"
+    "  float y = 1.1640625 * (texture2D (ytex, coord).a - 0.0625);"
+    "  float uvr = int (texture2D (utex, coord).r * 32);"
+    "  float uvg = int (texture2D (utex, coord).g * 64);"
+    "  float uvb = int (texture2D (utex, coord).b * 32);"
+    "  float tg = floor (uvg / 8.0);"
+    "  float u = (uvb + (uvg - tg * 8.0) * 32.0) / 256.0;"
+    "  float v = (uvr * 8.0 + tg) / 256.0;"
+    "  u -= 0.5;"
+    "  v -= 0.5;"
+    "  vec4 color;"
+    "  color.r = y + 1.59765625 * v;"
+    "  color.g = y - 0.390625 * u - 0.8125 * v;"
+    "  color.b = y + 2.015625 * u;"
+    "  color.a = 1.0;"
+    "  cogl_color_out = color;}";
+
 static gchar *yv12_to_rgba_shader =
     "uniform sampler2D ytex;"
     "uniform sampler2D utex;"
     "uniform sampler2D vtex;"
     "void main () {"
     "  vec2 coord = vec2(cogl_tex_coord_in[0]);"
-    "  float y = 1.1640625 * (texture2D (ytex, coord).g - 0.0625);"
-    "  float u = texture2D (utex, coord).g - 0.5;"
-    "  float v = texture2D (vtex, coord).g - 0.5;"
+    "  float y = 1.1640625 * (texture2D (ytex, coord).a - 0.0625);"
+    "  float v = texture2D (utex, coord).a - 0.5;"
+    "  float u = texture2D (vtex, coord).a - 0.5;"
+    "  vec4 color;"
+    "  color.r = y + 1.59765625 * v;"
+    "  color.g = y - 0.390625 * u - 0.8125 * v;"
+    "  color.b = y + 2.015625 * u;"
+    "  color.a = 1.0;"
+    "  cogl_color_out = color;}";
+
+static gchar *i420_to_rgba_shader =
+    "uniform sampler2D ytex;"
+    "uniform sampler2D utex;"
+    "uniform sampler2D vtex;"
+    "void main () {"
+    "  vec2 coord = vec2(cogl_tex_coord_in[0]);"
+    "  float y = 1.1640625 * (texture2D (ytex, coord).a - 0.0625);"
+    "  float u = texture2D (utex, coord).a - 0.5;"
+    "  float v = texture2D (vtex, coord).a - 0.5;"
     "  vec4 color;"
     "  color.r = y + 1.59765625 * v;"
     "  color.g = y - 0.390625 * u - 0.8125 * v;"
@@ -112,19 +149,28 @@ static gchar *yv12_to_rgba_shader =
 
 #define BASE_SINK_CAPS "{ AYUV," \
                        "YV12," \
+                       "NV12," \
                        "I420," \
                        "RGBA," \
                        "BGRA," \
                        "RGB," \
                        "BGR }"
 
+#define BASE_GL_SINK_CAPS "{ RGBA }"
+
+#define GL_SINK_CAPS GST_VIDEO_CAPS_MAKE_WITH_FEATURES(     \
+    GST_CAPS_FEATURE_META_GST_VIDEO_GL_TEXTURE_UPLOAD_META, \
+    BASE_GL_SINK_CAPS)
 
 #define SINK_CAPS GST_VIDEO_CAPS_MAKE(BASE_SINK_CAPS)
 
 static GstStaticPadTemplate sinktemplate_all = GST_STATIC_PAD_TEMPLATE ("sink",
     GST_PAD_SINK,
     GST_PAD_ALWAYS,
-    GST_STATIC_CAPS (SINK_CAPS));
+    GST_STATIC_CAPS (
+        GL_SINK_CAPS ";"
+        SINK_CAPS
+    ));
 
 GST_DEBUG_CATEGORY_STATIC (clutter_gst_video_sink_debug);
 #define GST_CAT_DEFAULT clutter_gst_video_sink_debug
@@ -144,8 +190,10 @@ typedef enum
   CLUTTER_GST_RGB24,
   CLUTTER_GST_AYUV,
   CLUTTER_GST_YV12,
+  CLUTTER_GST_NV12,
   CLUTTER_GST_I420,
-  CLUTTER_GST_SURFACE
+  CLUTTER_GST_SURFACE,
+  CLUTTER_GST_GL_TEXTURE_UPLOAD,
 } ClutterGstVideoFormat;
 
 /*
@@ -173,6 +221,7 @@ typedef struct _ClutterGstSource
   GstBuffer *buffer;
   gboolean has_new_caps;
   gboolean stage_lost;
+  gboolean has_gl_texture_upload_meta;
 } ClutterGstSource;
 
 /*
@@ -188,6 +237,7 @@ typedef struct _ClutterGstRenderer
   ClutterGstVideoFormat format; /* the format handled by this renderer */
   int flags;                    /* ClutterGstFeatures ORed flags */
   GstStaticCaps caps;           /* caps handled by the renderer */
+  gpointer context;             /* rendering context if any */
 
   void (*init) (ClutterGstVideoSink * sink);
   void (*deinit) (ClutterGstVideoSink * sink);
@@ -216,6 +266,10 @@ struct _ClutterGstVideoSinkPrivate
 
   GArray *signal_handler_ids;
 
+  GstVideoCropMeta crop_meta;
+  gboolean has_crop_meta;
+  gboolean crop_meta_has_changed;
+
 #ifdef HAVE_HW_DECODER_SUPPORT
   GstSurfaceConverter *converter;
 
@@ -343,6 +397,9 @@ clutter_gst_parse_caps (GstCaps * caps,
     case GST_VIDEO_FORMAT_YV12:
       format = CLUTTER_GST_YV12;
       break;
+    case GST_VIDEO_FORMAT_NV12:
+      format = CLUTTER_GST_NV12;
+      break;
     case GST_VIDEO_FORMAT_I420:
       format = CLUTTER_GST_I420;
       break;
@@ -373,6 +430,9 @@ clutter_gst_parse_caps (GstCaps * caps,
       goto unhandled_format;
   }
 
+  if (priv->source->has_gl_texture_upload_meta)
+    format = CLUTTER_GST_GL_TEXTURE_UPLOAD;
+
   /* find a renderer that can display our format */
   renderer = clutter_gst_find_renderer_by_format (sink, format);
 
@@ -473,6 +533,8 @@ static gboolean
 clutter_gst_source_dispatch (GSource * source,
     GSourceFunc callback, gpointer user_data)
 {
+  GstVideoCropMeta *crop_meta;
+  GstVideoGLTextureUploadMeta *upload_meta;
   ClutterGstSource *gst_source = (ClutterGstSource *) source;
   ClutterGstVideoSinkPrivate *priv = gst_source->sink->priv;
   GstBuffer *buffer;
@@ -481,6 +543,37 @@ clutter_gst_source_dispatch (GSource * source,
 
   g_mutex_lock (&gst_source->buffer_lock);
 
+#ifdef CLUTTER_COGL_HAS_GL
+  if (!gst_source->has_gl_texture_upload_meta &&
+      (upload_meta = gst_buffer_get_video_gl_texture_upload_meta (gst_source->buffer))) {
+    if (priv->renderer)
+      priv->renderer->deinit (gst_source->sink);
+
+    priv->renderer = clutter_gst_find_renderer_by_format (gst_source->sink,
+        CLUTTER_GST_GL_TEXTURE_UPLOAD);
+
+    gst_source->has_gl_texture_upload_meta = TRUE;
+  }
+#endif
+
+  crop_meta = gst_buffer_get_video_crop_meta (gst_source->buffer);
+  if (crop_meta) {
+    priv->has_crop_meta = TRUE;
+
+    if (priv->crop_meta.x      == crop_meta->x     &&
+        priv->crop_meta.y      == crop_meta->y     &&
+        priv->crop_meta.width  == crop_meta->width &&
+        priv->crop_meta.height == crop_meta->height) {
+      priv->crop_meta_has_changed = FALSE;
+    } else {
+      priv->crop_meta.x = crop_meta->x;
+      priv->crop_meta.y = crop_meta->y;
+      priv->crop_meta.width = crop_meta->width;
+      priv->crop_meta.height = crop_meta->height;
+      priv->crop_meta_has_changed = TRUE;
+    }
+  }
+
   if (G_UNLIKELY (gst_source->has_new_caps)) {
     GstCaps *caps =
         gst_pad_get_current_caps (GST_BASE_SINK_PAD ((GST_BASE_SINK
@@ -520,7 +613,12 @@ clutter_gst_source_dispatch (GSource * source,
       /* FIXME : We already call this above ? */
       if (!clutter_gst_parse_caps (caps, gst_source->sink, TRUE))
         goto negotiation_fail;
-      clutter_actor_set_size (stage, priv->info.width, priv->info.height);
+
+      if (priv->has_crop_meta)
+        clutter_actor_set_size (stage,
+            priv->crop_meta.width, priv->crop_meta.height);
+      else
+        clutter_actor_set_size (stage, priv->info.width, priv->info.height);
       clutter_actor_show (stage);
     } else {
       /* FIXME : We already call this above ? */
@@ -683,6 +781,29 @@ _create_cogl_program (const char *source)
   return program;
 }
 
+static CoglHandle
+_get_cached_cogl_program (const char *source)
+{
+  static GHashTable *program_cache = NULL;
+  CoglHandle handle;
+
+  if (G_UNLIKELY (program_cache == NULL)) {
+    program_cache = g_hash_table_new_full (g_str_hash, g_str_equal,
+                                           (GDestroyNotify) g_free,
+                                           (GDestroyNotify) cogl_handle_unref);
+  }
+
+  handle = (CoglHandle) g_hash_table_lookup (program_cache, (gpointer) source);
+  if (handle == COGL_INVALID_HANDLE) {
+    handle = _create_cogl_program (source);
+    g_hash_table_insert (program_cache,
+                         (gpointer) g_strdup (source),
+                         (gpointer) cogl_handle_ref (handle));
+  }
+
+  return handle;
+}
+
 static void
 _create_template_material (ClutterGstVideoSink * sink,
     const char *source, gboolean set_uniforms, int n_layers)
@@ -697,7 +818,7 @@ _create_template_material (ClutterGstVideoSink * sink,
   template = cogl_material_new ();
 
   if (source) {
-    CoglHandle program = _create_cogl_program (source);
+    CoglHandle program = _get_cached_cogl_program (source);
 
     if (set_uniforms) {
       unsigned int location;
@@ -719,7 +840,6 @@ _create_template_material (ClutterGstVideoSink * sink,
     }
 
     cogl_material_set_user_program (template, program);
-    cogl_handle_unref (program);
   }
 
   for (i = 0; i < n_layers; i++)
@@ -785,10 +905,13 @@ clutter_gst_rgb24_upload (ClutterGstVideoSink * sink, GstBuffer * buffer)
   if (!gst_video_frame_map (&frame, &priv->info, buffer, GST_MAP_READ))
     goto map_fail;
 
-  tex = cogl_texture_new_from_data (priv->info.width,
-      priv->info.height,
+  tex = cogl_texture_new_from_data (GST_VIDEO_FRAME_COMP_WIDTH (&frame, 0),
+      GST_VIDEO_FRAME_COMP_HEIGHT (&frame, 0),
       CLUTTER_GST_TEXTURE_FLAGS,
-      format, format, priv->info.stride[0], frame.data[0]);
+      format,
+      format,
+      GST_VIDEO_FRAME_PLANE_STRIDE (&frame, 0),
+      GST_VIDEO_FRAME_PLANE_DATA (&frame, 0));
 
   gst_video_frame_unmap (&frame);
 
@@ -809,6 +932,7 @@ static ClutterGstRenderer rgb24_renderer = {
   CLUTTER_GST_RGB24,
   0,
   GST_STATIC_CAPS (GST_VIDEO_CAPS_MAKE ("{ RGB, BGR }")),
+  NULL,
   clutter_gst_rgb_init,
   clutter_gst_dummy_deinit,
   clutter_gst_rgb24_upload,
@@ -834,10 +958,13 @@ clutter_gst_rgb32_upload (ClutterGstVideoSink * sink, GstBuffer * buffer)
   else
     format = COGL_PIXEL_FORMAT_RGBA_8888;
 
-  tex = cogl_texture_new_from_data (priv->info.width,
-      priv->info.height,
+  tex = cogl_texture_new_from_data (GST_VIDEO_FRAME_COMP_WIDTH (&frame, 0),
+      GST_VIDEO_FRAME_COMP_HEIGHT (&frame, 0),
       CLUTTER_GST_TEXTURE_FLAGS,
-      format, format, priv->info.stride[0], frame.data[0]);
+      format,
+      format,
+      GST_VIDEO_FRAME_PLANE_STRIDE (&frame, 0),
+      GST_VIDEO_FRAME_PLANE_DATA (&frame, 0));
 
   gst_video_frame_unmap (&frame);
 
@@ -858,6 +985,7 @@ static ClutterGstRenderer rgb32_renderer = {
   CLUTTER_GST_RGB32,
   0,
   GST_STATIC_CAPS (GST_VIDEO_CAPS_MAKE ("{ RGBA, BGRA }")),
+  NULL,
   clutter_gst_rgb_init,
   clutter_gst_dummy_deinit,
   clutter_gst_rgb32_upload,
@@ -872,34 +1000,28 @@ static ClutterGstRenderer rgb32_renderer = {
 static gboolean
 clutter_gst_yv12_upload (ClutterGstVideoSink * sink, GstBuffer * buffer)
 {
-  ClutterGstVideoSinkPrivate *priv = sink->priv;
-  CoglHandle y_tex, u_tex, v_tex;
+  int i;
+  CoglHandle texs[3];
   GstVideoFrame frame;
+  ClutterGstVideoSinkPrivate *priv = sink->priv;
 
   if (!gst_video_frame_map (&frame, &priv->info, buffer, GST_MAP_READ))
     goto no_map;
 
-  y_tex =
-      cogl_texture_new_from_data (GST_VIDEO_INFO_COMP_WIDTH (&priv->info, 0),
-      GST_VIDEO_INFO_COMP_HEIGHT (&priv->info, 0), CLUTTER_GST_TEXTURE_FLAGS,
-      COGL_PIXEL_FORMAT_G_8, COGL_PIXEL_FORMAT_G_8, priv->info.stride[0],
-      frame.data[0]);
-
-  u_tex =
-      cogl_texture_new_from_data (GST_VIDEO_INFO_COMP_WIDTH (&priv->info, 1),
-      GST_VIDEO_INFO_COMP_HEIGHT (&priv->info, 1), CLUTTER_GST_TEXTURE_FLAGS,
-      COGL_PIXEL_FORMAT_G_8, COGL_PIXEL_FORMAT_G_8, priv->info.stride[1],
-      frame.data[1]);
-
-  v_tex =
-      cogl_texture_new_from_data (GST_VIDEO_INFO_COMP_WIDTH (&priv->info, 2),
-      GST_VIDEO_INFO_COMP_HEIGHT (&priv->info, 2), CLUTTER_GST_TEXTURE_FLAGS,
-      COGL_PIXEL_FORMAT_G_8, COGL_PIXEL_FORMAT_G_8, priv->info.stride[2],
-      frame.data[2]);
+  for (i = 0; i < 3; i++) {
+    texs[i] =
+      cogl_texture_new_from_data (GST_VIDEO_FRAME_COMP_WIDTH (&frame, i),
+      GST_VIDEO_FRAME_COMP_HEIGHT (&frame, i),
+      CLUTTER_GST_TEXTURE_FLAGS,
+      COGL_PIXEL_FORMAT_A_8,
+      COGL_PIXEL_FORMAT_A_8,
+      GST_VIDEO_FRAME_PLANE_STRIDE (&frame, i),
+      GST_VIDEO_FRAME_PLANE_DATA (&frame, i));
+  }
 
   gst_video_frame_unmap (&frame);
 
-  _create_paint_material (sink, y_tex, u_tex, v_tex);
+  _create_paint_material (sink, texs[0], texs[1], texs[2]);
 
   return TRUE;
 
@@ -923,11 +1045,90 @@ static ClutterGstRenderer yv12_glsl_renderer = {
   CLUTTER_GST_YV12,
   CLUTTER_GST_GLSL | CLUTTER_GST_MULTI_TEXTURE,
   GST_STATIC_CAPS (GST_VIDEO_CAPS_MAKE ("YV12")),
+  NULL,
   clutter_gst_yv12_glsl_init,
   clutter_gst_dummy_deinit,
   clutter_gst_yv12_upload,
 };
 
+/*
+ * NV12
+ *
+ * 8 bit Y plane followed by interleaved U/V plane containing 8 bit 2x2 subsampled UV
+ */
+
+static gboolean
+clutter_gst_nv12_upload (ClutterGstVideoSink * sink, GstBuffer * buffer)
+{
+  ClutterGstVideoSinkPrivate *priv = sink->priv;
+  CoglMaterial *material;
+  CoglHandle y_tex, u_tex;
+  GstVideoFrame frame;
+
+  if (!gst_video_frame_map (&frame, &priv->info, buffer, GST_MAP_READ))
+    goto no_map;
+
+  y_tex =
+    cogl_texture_new_from_data (GST_VIDEO_FRAME_COMP_WIDTH (&frame, 0),
+    GST_VIDEO_FRAME_COMP_HEIGHT (&frame, 0),
+    CLUTTER_GST_TEXTURE_FLAGS,
+    COGL_PIXEL_FORMAT_A_8,
+    COGL_PIXEL_FORMAT_A_8,
+    GST_VIDEO_FRAME_PLANE_STRIDE (&frame, 0),
+    GST_VIDEO_FRAME_PLANE_DATA (&frame, 0));
+
+  u_tex =
+    cogl_texture_new_from_data (GST_VIDEO_FRAME_COMP_WIDTH (&frame, 1),
+    GST_VIDEO_FRAME_COMP_HEIGHT (&frame, 1),
+    CLUTTER_GST_TEXTURE_FLAGS,
+    COGL_PIXEL_FORMAT_RGB_565,
+    COGL_PIXEL_FORMAT_RGB_565,
+    GST_VIDEO_FRAME_PLANE_STRIDE (&frame, 1),
+    GST_VIDEO_FRAME_PLANE_DATA (&frame, 1));
+
+  gst_video_frame_unmap (&frame);
+
+  material = cogl_material_copy (priv->material_template);
+
+  cogl_material_set_layer (material, 0, y_tex);
+  cogl_material_set_layer (material, 1, u_tex);
+  cogl_material_set_layer_filters (material, 1,
+      COGL_MATERIAL_FILTER_NEAREST, COGL_MATERIAL_FILTER_NEAREST);
+
+  cogl_handle_unref (y_tex);
+  cogl_handle_unref (u_tex);
+
+  clutter_texture_set_cogl_material (priv->texture, material);
+  cogl_object_unref (material);
+
+  return TRUE;
+
+  /* ERRORS */
+no_map:
+  {
+    GST_ERROR_OBJECT (sink, "Could not map incoming video frame");
+    return FALSE;
+  }
+}
+
+static void
+clutter_gst_nv12_glsl_init (ClutterGstVideoSink * sink)
+{
+  _create_template_material (sink, nv12_to_rgba_shader, TRUE, 2);
+}
+
+
+static ClutterGstRenderer nv12_glsl_renderer = {
+  "NV12 glsl",
+  CLUTTER_GST_NV12,
+  CLUTTER_GST_GLSL | CLUTTER_GST_MULTI_TEXTURE,
+  GST_STATIC_CAPS (GST_VIDEO_CAPS_MAKE ("NV12")),
+  NULL,
+  clutter_gst_nv12_glsl_init,
+  clutter_gst_dummy_deinit,
+  clutter_gst_nv12_upload,
+};
+
 /*
  * YV12 (fragment program version)
  *
@@ -951,6 +1152,7 @@ static ClutterGstRenderer yv12_fp_renderer = {
   CLUTTER_GST_YV12,
   CLUTTER_GST_FP | CLUTTER_GST_MULTI_TEXTURE,
   GST_STATIC_CAPS (GST_VIDEO_CAPS_MAKE ("YV12")),
+  NULL,
   clutter_gst_yv12_fp_init,
   clutter_gst_dummy_deinit,
   clutter_gst_yv12_upload,
@@ -967,7 +1169,7 @@ static ClutterGstRenderer yv12_fp_renderer = {
 static void
 clutter_gst_i420_glsl_init (ClutterGstVideoSink * sink)
 {
-  _create_template_material (sink, yv12_to_rgba_shader, TRUE, 3);
+  _create_template_material (sink, i420_to_rgba_shader, TRUE, 3);
 }
 
 static ClutterGstRenderer i420_glsl_renderer = {
@@ -975,6 +1177,7 @@ static ClutterGstRenderer i420_glsl_renderer = {
   CLUTTER_GST_I420,
   CLUTTER_GST_GLSL | CLUTTER_GST_MULTI_TEXTURE,
   GST_STATIC_CAPS (GST_VIDEO_CAPS_MAKE ("I420")),
+  NULL,
   clutter_gst_i420_glsl_init,
   clutter_gst_dummy_deinit,
   clutter_gst_yv12_upload,
@@ -1004,6 +1207,7 @@ static ClutterGstRenderer i420_fp_renderer = {
   CLUTTER_GST_I420,
   CLUTTER_GST_FP | CLUTTER_GST_MULTI_TEXTURE,
   GST_STATIC_CAPS (GST_VIDEO_CAPS_MAKE ("I420")),
+  NULL,
   clutter_gst_i420_fp_init,
   clutter_gst_dummy_deinit,
   clutter_gst_yv12_upload,
@@ -1035,11 +1239,13 @@ clutter_gst_ayuv_upload (ClutterGstVideoSink * sink, GstBuffer * buffer)
     goto map_fail;
 
   tex =
-      cogl_texture_new_from_data (priv->info.width,
-      priv->info.height,
+      cogl_texture_new_from_data (GST_VIDEO_FRAME_COMP_WIDTH (&frame, 0),
+      GST_VIDEO_FRAME_COMP_HEIGHT (&frame, 0),
       CLUTTER_GST_TEXTURE_FLAGS,
       COGL_PIXEL_FORMAT_RGBA_8888,
-      COGL_PIXEL_FORMAT_RGBA_8888, priv->info.stride[0], frame.data[0]);
+      COGL_PIXEL_FORMAT_RGBA_8888,
+      GST_VIDEO_FRAME_PLANE_STRIDE (&frame, 0),
+      GST_VIDEO_FRAME_PLANE_DATA (&frame, 0));
 
   gst_video_frame_unmap (&frame);
 
@@ -1060,6 +1266,7 @@ static ClutterGstRenderer ayuv_glsl_renderer = {
   CLUTTER_GST_AYUV,
   CLUTTER_GST_GLSL,
   GST_STATIC_CAPS (GST_VIDEO_CAPS_MAKE ("AYUV")),
+  NULL,
   clutter_gst_ayuv_glsl_init,
   clutter_gst_dummy_deinit,
   clutter_gst_ayuv_upload,
@@ -1221,12 +1428,154 @@ static ClutterGstRenderer hw_renderer = {
   CLUTTER_GST_SURFACE,
   0,
   GST_STATIC_CAPS ("video/x-surface, opengl=true"),
+  NULL,
   clutter_gst_hw_init,
   clutter_gst_hw_deinit,
   clutter_gst_hw_upload,
 };
 #endif
 
+#ifdef CLUTTER_COGL_HAS_GL
+
+typedef struct {
+  gboolean is_initialized;
+} GLTextureUploadRendererContext;
+
+static void
+clutter_gst_gl_texture_upload_init (ClutterGstVideoSink * sink)
+{
+  ClutterGstRenderer *renderer = sink->priv->renderer;
+
+  if (renderer->context)
+    return;
+
+  renderer->context = g_new0 (GLTextureUploadRendererContext, 1);
+  if (!renderer->context) {
+    GST_ERROR ("Failed to allocate renderer context");
+  }
+}
+
+static void
+clutter_gst_gl_texture_upload_deinit (ClutterGstVideoSink * sink)
+{
+  ClutterGstRenderer *renderer = sink->priv->renderer;
+
+  if (!renderer->context)
+    return;
+
+  g_free (renderer->context);
+  renderer->context = NULL;
+}
+
+static gboolean
+clutter_gst_gl_texture_upload_init_texture (ClutterGstVideoSink * sink)
+{
+  CoglHandle material;
+  CoglTexture *tex = NULL, *crop_tex = NULL;
+  ClutterGstVideoSinkPrivate *priv = sink->priv;
+  GstVideoCropMeta *crop_meta = &priv->crop_meta;
+  ClutterGstRenderer *renderer = sink->priv->renderer;
+  GLTextureUploadRendererContext *context = renderer->context;
+
+  tex = cogl_texture_new_with_size (priv->info.width,
+      priv->info.height, CLUTTER_GST_TEXTURE_FLAGS, COGL_PIXEL_FORMAT_RGBA_8888);
+
+  if (!tex) {
+    GST_WARNING ("Couldn't create cogl texture");
+    return FALSE;
+  }
+
+  if (priv->has_crop_meta) {
+    crop_tex = cogl_texture_new_from_sub_texture (tex, crop_meta->x,
+        crop_meta->y, crop_meta->width, crop_meta->height);
+  }
+
+  material = cogl_material_new ();
+  if (!material) {
+    GST_WARNING ("Couldn't create cogl material");
+    return FALSE;
+  }
+
+  if (priv->has_crop_meta)
+    cogl_material_set_layer (material, 0, crop_tex);
+  else
+    cogl_material_set_layer (material, 0, tex);
+
+  clutter_texture_set_cogl_material (priv->texture, material);
+
+  cogl_object_unref (tex);
+  if (crop_tex)
+    cogl_object_unref (crop_tex);
+  cogl_object_unref (material);
+
+  context->is_initialized = TRUE;
+  return TRUE;
+}
+
+static gboolean
+clutter_gst_gl_texture_upload_upload (ClutterGstVideoSink * sink, GstBuffer * buffer)
+{
+  ClutterGstVideoSinkPrivate *priv = sink->priv;
+  ClutterGstRenderer *renderer = sink->priv->renderer;
+  GLTextureUploadRendererContext *context = renderer->context;
+  GstVideoGLTextureUploadMeta *upload_meta = NULL;
+  CoglTexture *tex;
+  guint gl_handle[4], gl_target[4];
+
+  if (!context) {
+    GST_WARNING ("Couldn't get the renderer context");
+    return FALSE;
+  }
+
+  if (!context->is_initialized || priv->crop_meta_has_changed) {
+    gboolean ret = clutter_gst_gl_texture_upload_init_texture (sink);
+    if (!ret)
+      return ret;
+  }
+
+  upload_meta = gst_buffer_get_video_gl_texture_upload_meta (buffer);
+  if (!upload_meta) {
+    GST_WARNING ("Buffer does not support GLTextureUploadMeta API");
+    return FALSE;
+  }
+
+  if (upload_meta->n_textures != 1 ||
+      upload_meta->texture_type[0] != GST_VIDEO_GL_TEXTURE_TYPE_RGBA) {
+    GST_WARNING ("clutter-video-sink only supports gl upload in a single RGBA texture");
+    return FALSE;
+  }
+
+  if (!(tex = clutter_texture_get_cogl_texture (priv->texture))) {
+    GST_WARNING ("Couldn't get Cogl texture");
+    return FALSE;
+  }
+
+  if (!cogl_texture_get_gl_texture (tex, gl_handle, gl_target)) {
+    GST_WARNING ("Couldn't get GL texture");
+    return FALSE;
+  }
+
+  if (!gst_video_gl_texture_upload_meta_upload (upload_meta, gl_handle)) {
+    GST_WARNING ("GL texture upload failed");
+    return FALSE;
+  }
+
+  clutter_actor_queue_redraw (CLUTTER_ACTOR (priv->texture));
+  return TRUE;
+}
+
+static ClutterGstRenderer gl_texture_upload_renderer = {
+    "GL Texture upload renderer",
+    CLUTTER_GST_GL_TEXTURE_UPLOAD,
+    0,
+    GST_STATIC_CAPS (GL_SINK_CAPS),
+    NULL,
+    clutter_gst_gl_texture_upload_init,
+    clutter_gst_gl_texture_upload_deinit,
+    clutter_gst_gl_texture_upload_upload,
+};
+#endif
+
 static GSList *
 clutter_gst_build_renderers_list (void)
 {
@@ -1241,6 +1590,7 @@ clutter_gst_build_renderers_list (void)
     &rgb24_renderer,
     &rgb32_renderer,
     &yv12_glsl_renderer,
+    &nv12_glsl_renderer,
     &i420_glsl_renderer,
 #ifdef CLUTTER_COGL_HAS_GL
     &yv12_fp_renderer,
@@ -1249,6 +1599,9 @@ clutter_gst_build_renderers_list (void)
     &ayuv_glsl_renderer,
 #ifdef HAVE_HW_DECODER_SUPPORT
     &hw_renderer,
+#endif
+#ifdef CLUTTER_COGL_HAS_GL
+    &gl_texture_upload_renderer,
 #endif
     NULL
   };
@@ -1466,6 +1819,11 @@ clutter_gst_video_sink_dispose (GObject * object)
   self = CLUTTER_GST_VIDEO_SINK (object);
   priv = self->priv;
 
+  if (priv->material_template != COGL_INVALID_HANDLE) {
+    cogl_object_unref (priv->material_template);
+    priv->material_template = COGL_INVALID_HANDLE;
+  }
+
   if (priv->renderer) {
     priv->renderer->deinit (self);
     priv->renderer = NULL;
@@ -1612,6 +1970,48 @@ clutter_gst_video_sink_stop (GstBaseSink * base_sink)
   return TRUE;
 }
 
+static gboolean
+clutter_gst_video_sink_propose_allocation (GstBaseSink * base_sink, GstQuery * query)
+{
+  gboolean need_pool = FALSE;
+  GstCaps * caps = NULL;
+
+  gst_query_parse_allocation (query, &caps, &need_pool);
+
+  gst_query_add_allocation_meta (query,
+      GST_VIDEO_META_API_TYPE, NULL);
+
+  gst_query_add_allocation_meta (query,
+      GST_VIDEO_GL_TEXTURE_UPLOAD_META_API_TYPE, NULL);
+
+  return TRUE;
+}
+
+static gboolean
+clutter_gst_video_sink_event (GstBaseSink * basesink, GstEvent * event)
+{
+  ClutterGstVideoSink *sink = CLUTTER_GST_VIDEO_SINK (basesink);
+  ClutterGstVideoSinkPrivate *priv = sink->priv;
+  ClutterGstSource *gst_source = priv->source;
+
+  switch (GST_EVENT_TYPE (event)) {
+    case GST_EVENT_FLUSH_START:
+      g_mutex_lock (&gst_source->buffer_lock);
+      if (gst_source->buffer) {
+        GST_DEBUG ("Freeing existing buffer %p", gst_source->buffer);
+        gst_buffer_unref (gst_source->buffer);
+        gst_source->buffer = NULL;
+      }
+      g_mutex_unlock (&gst_source->buffer_lock);
+      break;
+
+    default:
+      break;
+  }
+
+  return GST_BASE_SINK_CLASS (parent_class)->event (basesink, event);
+}
+
 static void
 clutter_gst_video_sink_class_init (ClutterGstVideoSinkClass * klass)
 {
@@ -1647,6 +2047,8 @@ clutter_gst_video_sink_class_init (ClutterGstVideoSinkClass * klass)
   gstbase_sink_class->stop = clutter_gst_video_sink_stop;
   gstbase_sink_class->set_caps = clutter_gst_video_sink_set_caps;
   gstbase_sink_class->get_caps = clutter_gst_video_sink_get_caps;
+  gstbase_sink_class->propose_allocation = clutter_gst_video_sink_propose_allocation;
+  gstbase_sink_class->event = clutter_gst_video_sink_event;
 
   /**
    * ClutterGstVideoSink:texture:
